package cc.fascinated.platform.impl;

import cc.fascinated.common.MathUtils;
import cc.fascinated.model.score.TotalScoresResponse;
import cc.fascinated.model.score.TrackedScore;
import cc.fascinated.model.token.ScoreSaberAccountToken;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.user.User;
import cc.fascinated.platform.CurvePoint;
import cc.fascinated.platform.Platform;
import cc.fascinated.services.QuestDBService;
import cc.fascinated.services.ScoreSaberService;
import cc.fascinated.services.TrackedScoreService;
import cc.fascinated.services.UserService;
import io.questdb.client.Sender;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@Log4j2
public class ScoreSaberPlatform extends Platform {
    /**
     * Delay in ms for requests per minute.
     */
    private static final long UPDATE_DELAY = 1000L / 150L; // 150 requests per minute

    /**
     * The base multiplier for stars.
     */
    private final double starMultiplier = 42.114296;

    /**
     * The ScoreSaber service to use
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    /**
     * The user service to use
     */
    @NonNull
    private final UserService userService;

    /**
     * The Influx service to use
     */
    @NonNull
    private final QuestDBService questDBService;

    /**
     * The tracked score service to use
     */
    @NonNull
    private final TrackedScoreService trackedScoreService;

    @Autowired
    public ScoreSaberPlatform(@NonNull ScoreSaberService scoreSaberService, @NonNull UserService userService, @NonNull QuestDBService questDBService,
                              @NonNull TrackedScoreService trackedScoreService) {
        super(Platforms.SCORESABER, 1, Map.of(
                1, new CurvePoint[]{
                        new CurvePoint(0, 0),
                        new CurvePoint(0.6, 0.18223233667439062),
                        new CurvePoint(0.65, 0.5866010012767576),
                        new CurvePoint(0.7, 0.6125565959114954),
                        new CurvePoint(0.75, 0.6451808210101443),
                        new CurvePoint(0.8, 0.6872268862950283),
                        new CurvePoint(0.825, 0.7150465663454271),
                        new CurvePoint(0.85, 0.7462290664143185),
                        new CurvePoint(0.875, 0.7816934560296046),
                        new CurvePoint(0.9, 0.825756123560842),
                        new CurvePoint(0.91, 0.8488375988124467),
                        new CurvePoint(0.92, 0.8728710341448851),
                        new CurvePoint(0.93, 0.9039994071865736),
                        new CurvePoint(0.94, 0.9417362980580238),
                        new CurvePoint(0.95, 1),
                        new CurvePoint(0.955, 1.0388633331418984),
                        new CurvePoint(0.96, 1.0871883573850478),
                        new CurvePoint(0.965, 1.1552120359501035),
                        new CurvePoint(0.97, 1.2485807759957321),
                        new CurvePoint(0.9725, 1.3090333065057616),
                        new CurvePoint(0.975, 1.3807102743105126),
                        new CurvePoint(0.9775, 1.4664726399289512),
                        new CurvePoint(0.98, 1.5702410055532239),
                        new CurvePoint(0.9825, 1.697536248647543),
                        new CurvePoint(0.985, 1.8563887693647105),
                        new CurvePoint(0.9875, 2.058947159052738),
                        new CurvePoint(0.99, 2.324506282149922),
                        new CurvePoint(0.99125, 2.4902905794106913),
                        new CurvePoint(0.9925, 2.685667856592722),
                        new CurvePoint(0.99375, 2.9190155639254955),
                        new CurvePoint(0.995, 3.2022017597337955),
                        new CurvePoint(0.99625, 3.5526145337555373),
                        new CurvePoint(0.9975, 3.996793606763322),
                        new CurvePoint(0.99825, 4.325027383589547),
                        new CurvePoint(0.999, 4.715470646416203),
                        new CurvePoint(0.9995, 5.019543595874787),
                        new CurvePoint(1, 5.367394282890631),
                }
        ));
        this.scoreSaberService = scoreSaberService;
        this.userService = userService;
        this.questDBService = questDBService;
        this.trackedScoreService = trackedScoreService;
    }

    /**
     * Gets the modifier for the given accuracy.
     *
     * @param accuracy The accuracy.
     * @return The modifier.
     */
    public double getModifier(double accuracy) {
        accuracy = MathUtils.clamp(accuracy, 0, 100) / 100;
        CurvePoint[] curve = this.getCurve(this.getCurrentCurveVersion());
        if (accuracy <= 0) {
            return 0;
        }

        if (accuracy >= 1) {
            return curve[curve.length - 1].getMultiplier();
        }

        for (int i = 0; i < curve.length - 1; i++) {
            CurvePoint point = curve[i];
            CurvePoint nextPoint = curve[i + 1];
            if (accuracy >= point.getAcc() && accuracy <= nextPoint.getAcc()) {
                return MathUtils.lerp(point.getMultiplier(), nextPoint.getMultiplier(), (accuracy - point.getAcc()) / (nextPoint.getAcc() - point.getAcc()));
            }
        }

        return 0;
    }

    @Override
    public double getPp(double stars, double accuracy) {
        if (accuracy <= 1) { // Convert the accuracy to a percentage
            accuracy *= 100;
        }
        double pp = stars * this.starMultiplier;
        return this.getModifier(accuracy) * pp;
    }

    @Override
    public void updatePlayers() {
        for (User user : this.userService.getUsers()) {
            if (!user.isHasLoggedIn()) { // Check if the user has linked their account
                continue;
            }
            ScoreSaberAccountToken account = scoreSaberService.getAccount(user); // Get the account from the ScoreSaber API

            try (Sender sender = questDBService.getSender()) {
                sender.table("player")
                        .symbol("platform", this.getPlatform().getPlatformName())
                        .symbol("user_id", user.getId().toString())
                        .doubleColumn("pp", account.getPp())
                        .longColumn("rank", account.getRank())
                        .longColumn("country_rank", account.getCountryRank())
                        .longColumn("total_score", account.getScoreStats().getTotalScore())
                        .longColumn("total_ranked_score", account.getScoreStats().getTotalRankedScore())
                        .doubleColumn("average_ranked_accuracy", account.getScoreStats().getAverageRankedAccuracy())
                        .longColumn("total_play_count", account.getScoreStats().getTotalPlayCount())
                        .longColumn("ranked_play_count", account.getScoreStats().getRankedPlayCount())
                        .atNow();
            }
        }
    }

    @Override
    public void updateMetrics() {
        try (Sender sender = questDBService.getSender()) {
            TotalScoresResponse totalScores = trackedScoreService.getTotalScores(this.getPlatform());
            sender.table("metrics")
                    .symbol("platform", this.getPlatform().getPlatformName())
                    .longColumn("total_scores", totalScores.getTotalScores())
                    .longColumn("total_ranked_scores", totalScores.getTotalRankedScores())
                    .atNow();
        }
    }

    @Override
    public void updateLeaderboards() {
        List<TrackedScore> scores = this.trackedScoreService.getTrackedScores(this.getPlatform(), false);
        List<String> uniqueLeaderboardIds = scores.stream().map(TrackedScore::getLeaderboardId).distinct().toList();

        log.info("Updating {} leaderboards for platform '{}'",
                uniqueLeaderboardIds.size(),
                this.getPlatform().getPlatformName()
        );

        // Update the leaderboards
        int finished = 0;
        for (String id : uniqueLeaderboardIds) {
            try {
                ScoreSaberLeaderboardToken leaderboard = this.scoreSaberService.getLeaderboard(id, true); // Update the cached leaderboard
                if (leaderboard == null) {
                    log.warn("Failed to update leaderboard '{}' for platform '{}'", id, this.getPlatform().getPlatformName());
                    continue;
                }
                if (leaderboard.getStars() == 0) { // Check if the leaderboard is ranked
                    log.warn("Leaderboard '{}' for platform '{}' is not ranked, skipping updating the score pp values",
                            id,
                            this.getPlatform().getPlatformName()
                    );
                    continue;
                }

                List<TrackedScore> toUpdate = scores.stream().filter(score -> {
                    if (!score.getLeaderboardId().equals(id)) { // Check if the leaderboard ID matches
                        return false;
                    }
                    double oldPp = score.getPp() == null ? 0 : score.getPp();
                    double pp = this.getPp(leaderboard.getStars(), score.getAccuracy());
                    log.info("Leaderboard: {}, Old PP: {}, New PP: {}", leaderboard.getSongName(), oldPp, pp);
                    return pp != oldPp; // Check if the pp has changed
                }).toList();

                for (TrackedScore score : toUpdate) { // Update the scores
                    double pp = this.getPp(leaderboard.getStars(), score.getAccuracy());
                    score.setPp(pp);
                }

                if (!toUpdate.isEmpty()) { // Save the scores
                    this.trackedScoreService.updateScores(toUpdate.toArray(TrackedScore[]::new));
                }
                finished++;

                log.info("Updated leaderboard '{}' for platform '{}', changed {} scores. ({}/{})",
                        id,
                        this.getPlatform().getPlatformName(),
                        toUpdate.size(),
                        finished,
                        uniqueLeaderboardIds.size()
                );

                // Sleep to prevent rate limiting
                try {
                    Thread.sleep(UPDATE_DELAY);
                } catch (InterruptedException e) {
                    log.error("Failed to sleep for rate limit reset", e);
                }
            } catch (Exception ex) {
                log.error("An error occurred while updating leaderboard '{}'", id, ex);
            }
        }
    }
}
