package cc.fascinated.repository.couchdb;

import cc.fascinated.model.score.TrackedScore;
import jakarta.transaction.Transactional;
import org.springframework.data.jpa.repository.Modifying;
import org.springframework.data.jpa.repository.Query;
import org.springframework.data.repository.CrudRepository;
import org.springframework.data.repository.query.Param;

import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
public interface TrackedScoreRepository extends CrudRepository<TrackedScore, String> {
    /**
     * Ensures that the deduplication of the scores is done.
     */
    @Modifying @Transactional
    @Query(value = "ALTER TABLE score DEDUP ENABLE UPSERT KEYS(timestamp, score_id)", nativeQuery = true)
    void ensureDeduplication();

    /**
     * Updates the pp of a score.
     *
     * @param scoreId the ID of the score
     * @param pp the new pp of the score
     */
    @Modifying @Transactional
    @Query(value = "UPDATE score SET pp = :pp WHERE score_id = :scoreId", nativeQuery = true)
    void updateScorePp(@Param("scoreId") String scoreId, @Param("pp") double pp);

    /**
     * Gets a list of top tracked scores
     * sorted by pp from the platform
     *
     * @param platform the platform to get the scores from
     * @param amount the amount of scores to get
     * @return the scores
     */
    @Query(value = "SELECT * FROM score WHERE platform = :platform AND pp > 0 ORDER BY pp DESC LIMIT :amount", nativeQuery = true)
    List<TrackedScore> findTopRankedScores(@Param("platform") String platform, @Param("amount") int amount);

    /**
     * Gets all tracked scores from a platform.
     *
     * @param platform the platform to get the scores from
     * @return the scores
     */
    @Query(value = "SELECT * FROM score WHERE platform = :platform", nativeQuery = true)
    List<TrackedScore> findAllByPlatform(String platform);

    /**
     * Gets all tracked scores from a platform.
     *
     * @param platform the platform to get the scores from
     * @return the scores
     */
    @Query(value = "SELECT * FROM score WHERE platform = :platform AND pp > 0", nativeQuery = true)
    List<TrackedScore> findAllByPlatformRankedOnly(String platform);

    /**
     * Gets the total amount of scores
     * for a platform.
     *
     * @param platform the platform to get the scores from
     * @return the total amount of scores for the platform
     */
    @Query(value = "SELECT COUNT(*) FROM score WHERE platform = :platform", nativeQuery = true)
    int countTotalScores(@Param("platform") String platform);

    /**
     * Gets the total amount of ranked scores
     * for a platform.
     *
     * @param platform the platform to get the scores from
     * @return the total amount of ranked scores for the platform
     */
    @Query(value = "SELECT COUNT(*) FROM score WHERE platform = :platform AND pp > 0", nativeQuery = true)
    int countTotalRankedScores(@Param("platform") String platform);

    /**
     * Gets all scores for a platform.
     *
     * @param platform the platform to get the scores from
     * @return the scores
     */
    @Query(value = "SELECT COUNT(*) FROM score WHERE platform = :platform AND pp > :pp", nativeQuery = true)
    int getScoreCountOverPpThreshold(@Param("platform") String platform, @Param("pp") double pp);
}